/*
 * Name:    d_mardtb_status.c
 *
 * Purpose: MX driver for the MarUSA Desktop Beamline's 'status_dump' command.
 *
 *          The driver reads out the value of a particular parameter.  The
 *          most interesting parameters are parameter 136 and 137, which
 *          are the readouts for the goniostat ion chambers.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_analog_input.h"
#include "i_mardtb.h"
#include "d_mardtb_status.h"

MX_RECORD_FUNCTION_LIST mxd_mardtb_status_record_function_list = {
	NULL,
	mxd_mardtb_status_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

MX_ANALOG_INPUT_FUNCTION_LIST mxd_mardtb_status_analog_input_function_list = {
	mxd_mardtb_status_read
};

MX_RECORD_FIELD_DEFAULTS mxd_mardtb_status_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_LONG_ANALOG_INPUT_STANDARD_FIELDS,
	MX_ANALOG_INPUT_STANDARD_FIELDS,
	MXD_MARDTB_STATUS_STANDARD_FIELDS
};

long mxd_mardtb_status_num_record_fields
		= sizeof( mxd_mardtb_status_record_field_defaults )
			/ sizeof( mxd_mardtb_status_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_mardtb_status_rfield_def_ptr
			= &mxd_mardtb_status_record_field_defaults[0];

#define MARDTB_DEBUG	FALSE

/* ===== */

MX_EXPORT mx_status_type
mxd_mardtb_status_create_record_structures( MX_RECORD *record )
{
        const char fname[] = "mxd_mardtb_status_create_record_structures()";

        MX_ANALOG_INPUT *analog_input;
        MX_MARDTB_STATUS *mardtb_status;

        /* Allocate memory for the necessary structures. */

        analog_input = (MX_ANALOG_INPUT *) malloc(sizeof(MX_ANALOG_INPUT));

        if ( analog_input == (MX_ANALOG_INPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ANALOG_INPUT structure." );
        }

        mardtb_status = (MX_MARDTB_STATUS *)
				malloc( sizeof(MX_MARDTB_STATUS) );

        if ( mardtb_status == (MX_MARDTB_STATUS *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_MARDTB_STATUS structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = analog_input;
        record->record_type_struct = mardtb_status;
        record->class_specific_function_list
			= &mxd_mardtb_status_analog_input_function_list;

        analog_input->record = record;

	/* Raw analog input values are stored as longs. */

	analog_input->subclass = MXT_AIN_LONG;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mardtb_status_read( MX_ANALOG_INPUT *ainput )
{
	const char fname[] = "mxd_mardtb_status_read()";

	MX_MARDTB_STATUS *mardtb_status;
	MX_MARDTB *mardtb;
	MX_RECORD *mardtb_record;
	char command[80];
	char response[100];
	char *ptr;
	long raw_value;
	int num_items, move_in_progress;
	mx_status_type mx_status;

	mardtb_status = (MX_MARDTB_STATUS *)
				ainput->record->record_type_struct;

	if ( mardtb_status == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MARDTB_STATUS pointer for analog input '%s' is NULL",
			ainput->record->name );
	}

	mardtb_record = mardtb_status->mardtb_record;

	if ( mardtb_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mardtb_record pointer for analog input '%s' is NULL.",
			ainput->record->name );
	}

	mardtb = (MX_MARDTB *) mardtb_record->record_type_struct;

	if ( mardtb == (MX_MARDTB *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MARDTB pointer for Mar DTB record '%s' is NULL.",
			mardtb_record->name );
	}

	mx_status = mxi_mardtb_check_for_move_in_progress( mardtb,
							&move_in_progress );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( move_in_progress ) {
		mx_warning( "Mar DTB motor '%s' is currently in motion, "
		"so we cannot read the current state of analog input '%s'.  "
		"The cached value (%g) will be used instead.",
			mardtb->currently_active_record->name,
			ainput->record->name,
			ainput->value );

		return MX_SUCCESSFUL_RESULT;
	}

	sprintf( command, "status_dump 1, 1, %d",
				mardtb_status->parameter_number );

	mx_status = mxi_mardtb_command( mardtb, command,
				response, sizeof response, MARDTB_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Skip over the first eight characters, since this part of the
	 * response varies as to how many whitespace delimited fields
	 * there are.
	 */

	ptr = response + 8;

	num_items = sscanf( ptr, "%ld", &raw_value );

	if ( num_items != 1 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The response to the status dump command '%s' was not in a "
		"recognizable format.  Response = '%s'",
			command, response );
	}

	ainput->raw_value.long_value = raw_value;

	return mx_status;
}

