/*
 * Name:    d_mdrive.c 
 *
 * Purpose: MX motor driver to control Intelligent Motion Systems MDrive
 *          motor controllers.
 *
 * WARNING: This driver has not yet been tested with a real MDrive controller
 *          so it is almost certain not to work as is.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MDRIVE_DEBUG	TRUE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "mx_unistd.h"

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_ascii.h"
#include "mx_motor.h"
#include "mx_rs232.h"
#include "d_mdrive.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_mdrive_record_function_list = {
	NULL,
	mxd_mdrive_create_record_structures,
	mxd_mdrive_finish_record_initialization,
	NULL,
	mxd_mdrive_print_motor_structure,
	NULL,
	NULL,
	mxd_mdrive_open,
	NULL,
	NULL,
	mxd_mdrive_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_mdrive_motor_function_list = {
	NULL,
	mxd_mdrive_move_absolute,
	mxd_mdrive_get_position,
	mxd_mdrive_set_position,
	mxd_mdrive_soft_abort,
	mxd_mdrive_immediate_abort,
	NULL,
	NULL,
	mxd_mdrive_find_home_position,
	mxd_mdrive_constant_velocity_move,
	mxd_mdrive_get_parameter,
	mxd_mdrive_set_parameter,
	NULL,
	mxd_mdrive_get_status
};

/* MDrive MDRIVE motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_mdrive_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_MDRIVE_STANDARD_FIELDS
};

long mxd_mdrive_num_record_fields
		= sizeof( mxd_mdrive_record_field_defaults )
			/ sizeof( mxd_mdrive_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_mdrive_rfield_def_ptr
			= &mxd_mdrive_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_mdrive_get_pointers( MX_MOTOR *motor,
			MX_MDRIVE **mdrive,
			const char *calling_fname )
{
	static const char fname[] = "mxd_mdrive_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mdrive == (MX_MDRIVE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MDRIVE pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*mdrive = (MX_MDRIVE *) (motor->record->record_type_struct);

	if ( *mdrive == (MX_MDRIVE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MDRIVE pointer for record '%s' is NULL.",
			motor->record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type mxd_mdrive_handle_error( MX_MDRIVE *mdrive,
						char *command,
						char *error_message );

/* === */

MX_EXPORT mx_status_type
mxd_mdrive_create_record_structures( MX_RECORD *record )
{
	static const char fname[] = "mxd_mdrive_create_record_structures()";

	MX_MOTOR *motor;
	MX_MDRIVE *mdrive;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for an MX_MOTOR structure for record '%s'.",
			record->name );
	}

	mdrive = (MX_MDRIVE *) malloc( sizeof(MX_MDRIVE) );

	if ( mdrive == (MX_MDRIVE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for an MX_MDRIVE structure for record '%s'.",
			record->name );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = mdrive;
	record->class_specific_function_list
				= &mxd_mdrive_motor_function_list;

	motor->record = record;
	mdrive->record = record;

	/* An MDrive motor is treated as an stepper motor. */

	motor->subclass = MXC_MTR_STEPPER;

	/* The MDRIVE reports accelerations in steps/sec**2 */

	motor->acceleration_type = MXF_MTR_ACCEL_RATE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mdrive_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type mx_status;

	mx_status = mx_motor_finish_record_initialization( record );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_print_motor_structure( FILE *file, MX_RECORD *record )
{
	static const char fname[] = "mxd_mdrive_print_motor_structure()";

	MX_MOTOR *motor;
	MX_MDRIVE *mdrive;
	double position, move_deadband;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	fprintf(file, "  Motor type            = MDRIVE\n\n");
	fprintf(file, "  name                  = %s\n", record->name);
	fprintf(file, "  rs232                 = %s\n",
						mdrive->rs232_record->name);

	fprintf(file, "  axis name             = %d\n", mdrive->axis_name);
	fprintf(file, "  negative limit switch = %d\n",
						mdrive->negative_limit_switch);
	fprintf(file, "  positive limit switch = %d\n",
						mdrive->positive_limit_switch);
	fprintf(file, "  home switch           = %d\n",
						mdrive->home_switch);

	mx_status = mx_motor_get_position( record, &position );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read the position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position            = %g %s  (%ld steps)\n",
		motor->position, motor->units,
		motor->raw_position.stepper );
	fprintf(file, "  scale               = %g %s per step.\n",
		motor->scale, motor->units);
	fprintf(file, "  offset              = %g %s.\n",
		motor->offset, motor->units);

	fprintf(file, "  backlash            = %g %s  (%ld steps)\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.stepper);

	fprintf(file, "  negative limit      = %g %s  (%ld steps)\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.stepper );

	fprintf(file, "  positive limit      = %g %s  (%ld steps)\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.stepper );

	move_deadband = motor->scale * (double)motor->raw_move_deadband.stepper;

	fprintf(file, "  move deadband       = %g %s  (%ld steps)\n\n",
		move_deadband, motor->units,
		motor->raw_move_deadband.stepper );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mdrive_open( MX_RECORD *record )
{
	static const char fname[] = "mxd_mdrive_open()";

	MX_MOTOR *motor;
	MX_MDRIVE *mdrive;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Verify the line terminators. */

	mx_status = mx_rs232_verify_configuration( mdrive->rs232_record,
			(long) MXF_232_DONT_CARE, (int) MXF_232_DONT_CARE,
			(char) MXF_232_DONT_CARE, (int) MXF_232_DONT_CARE,
			(char) MXF_232_DONT_CARE,
			0x0d0a, 0x0d );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if 0
	{
		double delay_seconds;

		/* Server minimum delay time between commands. */

		delay_seconds = 0.01;

		record->event_interval
			= mx_convert_seconds_to_clock_ticks( delay_seconds );
	}
#endif

	mx_status = mxd_mdrive_resynchronize( record );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_resynchronize( MX_RECORD *record )
{
	static const char fname[] = "mxd_mdrive_resynchronize()";

	MX_MOTOR *motor;
	MX_MDRIVE *mdrive;
	char response[80];
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Discard any unread RS-232 input or unwritten RS-232 output. */

	mx_status = mx_rs232_discard_unwritten_output( mdrive->rs232_record,
							MDRIVE_DEBUG );

	if ( (mx_status.code != MXE_SUCCESS)
	  && (mx_status.code != MXE_UNSUPPORTED) )
		return mx_status;

	mx_status = mx_rs232_discard_unread_input( mdrive->rs232_record,
							MDRIVE_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Verify that the controller is there by asking for its
	 * firmware version.
	 */

	mx_status = mxd_mdrive_command( mdrive, "VR",
				response, sizeof(response), MDRIVE_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Enable the drive. */

	mx_status = mxd_mdrive_command( mdrive, "DE=1",
						NULL, 0, MDRIVE_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_move_absolute( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_move_absolute()";

	MX_MDRIVE *mdrive;
	char command[20];
	long motor_steps;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor_steps = motor->raw_destination.stepper;

	/* Format the move command and send it. */

	sprintf( command, "MA %ld", motor_steps );

	mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_get_position( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_get_position()";

	MX_MDRIVE *mdrive;
	char response[30];
	int num_items;
	long motor_steps;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_mdrive_command( mdrive, "PR P",
			response, sizeof response, MDRIVE_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%ld", &motor_steps );

	if ( num_items != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Motor '%s' returned an unparseable position value '%s'.",
			motor->record->name, response );
	}

	motor->raw_position.stepper = motor_steps;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mdrive_set_position( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_set_position()";

	MX_MDRIVE *mdrive;
	char command[40];
	long motor_steps;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor_steps = motor->raw_set_position.stepper;

	sprintf( command, "P %ld", motor_steps );

	mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_soft_abort( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_soft_abort()";

	MX_MDRIVE *mdrive;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* I don't see an obvious 'stop' command in the manual, so my best
	 * guess is to disable the drive.
	 */

	mx_status = mxd_mdrive_command( mdrive, "DE=0",
						NULL, 0, MDRIVE_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Wait a little while for the controller to react to the command. */

	mx_msleep(1000);

	/* Reenable the drive again. */

	mx_status = mxd_mdrive_command( mdrive, "DE=1",
						NULL, 0, MDRIVE_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_immediate_abort( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_immediate_abort()";

	MX_MDRIVE *mdrive;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* I don't see an obvious 'stop' command in the manual, so my best
	 * guess is to disable the drive.
	 */

	mx_status = mxd_mdrive_command( mdrive, "DE=0",
						NULL, 0, MDRIVE_DEBUG );

	/* The drive is left disabled. */

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_find_home_position( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_find_home_position()";

	MX_MDRIVE *mdrive;
	char command[20];
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( motor->home_search >= 0 ) {
		strcpy( command, "HM=3" );
	} else {
		strcpy( command, "HM=1" );
	}
		
	mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_constant_velocity_move( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_constant_velocity_move()";

	MX_MDRIVE *mdrive;
	char command[20];
	double speed;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_motor_get_speed( motor->record, &speed );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( motor->constant_velocity_move >= 0 ) {
		sprintf( command, "SL=%ld",
				mx_round( motor->raw_speed ) );
	} else {
		sprintf( command, "SL=-%ld",
				mx_round( motor->raw_speed ) );
	}

	mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_get_parameter( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_get_parameter()";

	MX_MDRIVE *mdrive;
	char response[80];
	int num_items;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		mx_status = mxd_mdrive_command( mdrive, "PR VM",
				response, sizeof(response), MDRIVE_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		num_items = sscanf( response, "%lg", &(motor->raw_speed) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"Unable to parse response to a 'PR VM' command "
			"for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}
		break;
	case MXLV_MTR_BASE_SPEED:
		mx_status = mxd_mdrive_command( mdrive, "PR VI",
				response, sizeof(response), MDRIVE_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		num_items = sscanf( response, "%lg", &(motor->raw_base_speed) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"Unable to parse response to a 'PR VI' command "
			"for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}
		break;
	case MXLV_MTR_RAW_ACCELERATION_PARAMETERS:
		mx_status = mxd_mdrive_command( mdrive, "PR A",
				response, sizeof(response), MDRIVE_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		num_items = sscanf( response, "%lg",
				&(motor->raw_acceleration_parameters[0]) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"Unable to parse response to a 'PR A' command "
			"for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

		mx_status = mxd_mdrive_command( mdrive, "PR D",
				response, sizeof(response), MDRIVE_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		num_items = sscanf( response, "%lg",
				&(motor->raw_acceleration_parameters[1]) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"Unable to parse response to a 'PR D' command "
			"for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

		motor->raw_acceleration_parameters[2] = 0.0;
		motor->raw_acceleration_parameters[3] = 0.0;
		break;

	case MXLV_MTR_MAXIMUM_SPEED:
		motor->raw_maximum_speed = 5000000; /* steps per second */
		break;

	default:
		return mx_motor_default_get_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mdrive_set_parameter( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_set_parameter()";

	MX_MDRIVE *mdrive;
	char command[20];
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( command, "VM=%ld", mx_round( motor->raw_speed ) );

		mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );
		break;

	case MXLV_MTR_BASE_SPEED:
		sprintf( command, "VI=%ld", mx_round( motor->raw_base_speed ) );

		mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );
		break;

	case MXLV_MTR_RAW_ACCELERATION_PARAMETERS:
		sprintf( command, "A=%ld",
			mx_round( motor->raw_acceleration_parameters[0] ));

		mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( command, "D=%ld",
			mx_round( motor->raw_acceleration_parameters[1] ));

		mx_status = mxd_mdrive_command( mdrive, command,
						NULL, 0, MDRIVE_DEBUG );
		break;

	default:
		return mx_motor_default_set_parameter_handler( motor );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mdrive_get_status( MX_MOTOR *motor )
{
	static const char fname[] = "mxd_mdrive_get_status()";

	MX_MDRIVE *mdrive;
	char response[80];
	int num_items, move_status, stall_status;
	mx_status_type mx_status;

	mx_status = mxd_mdrive_get_pointers( motor, &mdrive, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* This routine is not finished.  I need to spend some time with
	 * a real controller to determine what the right things to do are.
	 */

	motor->status = 0;

	/* Get the move status.  I am guessing as to what the response
	 * should look like.
	 */

	mx_status = mxd_mdrive_command( mdrive, "PR MV",
			response, sizeof(response), MDRIVE_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d", &move_status );

	if ( num_items != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Unable to parse response to a 'PR MV' command "
		"for motor '%s'.  Response = '%s'",
			motor->record->name, response );
	}

	if ( move_status ) {
		motor->status |= MXSF_MTR_IS_BUSY;
	}

	/* See if the motor is stalled. */

	mx_status = mxd_mdrive_command( mdrive, "PR ST",
			response, sizeof(response), MDRIVE_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d", &stall_status );

	if ( num_items != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Unable to parse response to a 'PR MV' command "
		"for motor '%s'.  Response = '%s'",
			motor->record->name, response );
	}

	if ( stall_status ) {
		motor->status |= MXSF_MTR_FOLLOWING_ERROR;
	}

	return mx_status;
}

/* === Extra functions for the use of this driver. === */

MX_EXPORT mx_status_type
mxd_mdrive_command( MX_MDRIVE *mdrive, char *command, char *response,
		int response_buffer_length, int debug_flag )
{
	static const char fname[] = "mxd_mdrive_command()";

	char prefix[20], local_response_buffer[100];
	char *ptr, *response_ptr;
	int response_prefix, prefix_separator;
	int buffer_length, prefix_length, chars_to_copy;
	int command_length, response_length;
	mx_status_type mx_status;

	if ( mdrive == (MX_MDRIVE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MDRIVE pointer passed was NULL." );
	}
	if ( command == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The 'command' pointer passed was NULL." );
	}

	/* Send the axis name prefix. */

	sprintf( prefix, "%c", mdrive->axis_name );

	if ( debug_flag ) {
		MX_DEBUG(-2, ("mxd_mdrive_command: command = '%s%s'",
					prefix, command));
	}

	prefix_length = strlen( prefix );

	command_length = prefix_length + strlen( command );

	mx_status = mx_rs232_write( mdrive->rs232_record,
				prefix, prefix_length,
				NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the command string. */

	mx_status = mx_rs232_putline( mdrive->rs232_record,
					command, NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Get the response. */

	if ( response == NULL ) {
		response_ptr = local_response_buffer;
		buffer_length = sizeof( local_response_buffer ) - 1;
	} else {
		response_ptr = response;
		buffer_length = response_buffer_length;
	}

	mx_status = mx_rs232_getline( mdrive->rs232_record,
					response_ptr, buffer_length, NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The PM600 echoes back the transmitted characters.  We must discard
	 * 'command_length + 1' received characters, where the extra character
	 * is the carriage return appended to the end of the string.
	 */

	response_ptr += command_length + 1;

	/* Delete any trailing line terminators. */

	response_length = strlen( response_ptr );

	if ( response_ptr[response_length - 1] == MX_LF ) {
		response_ptr[ response_length - 1 ] = '\0';

		if ( response_ptr[ response_length - 2 ] == MX_CR ) {
			response_ptr[ response_length - 2 ] = '\0';
		}
	}

#if 0
	response_length = strlen( response_ptr );

	{
		int i;

		for ( i = 0; i < response_length; i++ ) {
			MX_DEBUG(-2,("%s: response_ptr[%d] = %#x '%c'",
				fname, i, response_ptr[i], response_ptr[i]));
		}
	}
#endif

	if ( debug_flag ) {
		MX_DEBUG(-2, ("mxd_mdrive_command: response_ptr = '%s'",
					response_ptr ));
	}

	/* Find the prefix separator in the response. */

	prefix_separator = ':';

	ptr = strchr( response_ptr, prefix_separator );

	if ( ptr == NULL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Prefix separator '%c' not seen in response from motor '%s'.  "
		"Response = '%s'",
		prefix_separator, mdrive->record->name, response_ptr );
	}

	/* Null terminate the prefix. */

	*ptr = '\0';

	/* See if the response prefix matches the original prefix. */

	response_prefix = response_ptr[0];

	if ( response_prefix != mdrive->axis_name ) {
		*ptr = prefix_separator;

		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The axis name '%c' in the response '%s' for motor '%s' does not "
	"match the expected axis name of '%c'.",
			response_prefix, response_ptr,
			mdrive->record->name,
			mdrive->axis_name );
	}

	ptr++;

	/* Check to see if the MDrive responded with an error message. */

	if ( *ptr == '!' ) {
		mx_status = mxd_mdrive_handle_error( mdrive, command, ptr );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* If the response pointer passed was not null, move the body
	 * of the response to the beginning of the response buffer.
	 */

	if ( response != NULL ) {
		chars_to_copy = strlen( ptr );

		if ( chars_to_copy > response_buffer_length - 1 ) {
			chars_to_copy = response_buffer_length - 1;
		}

		memmove( response, ptr, chars_to_copy );

		response[chars_to_copy] = '\0';
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_mdrive_handle_error( MX_MDRIVE *mdrive, char *command, char *error_message )
{
	strcpy( error_message, "No error" );

	return MX_SUCCESSFUL_RESULT;
}

