/*
 * Name:    d_network_doutput.c
 *
 * Purpose: MX network digital output device driver.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001, 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_net.h"
#include "mx_digital_output.h"
#include "d_network_doutput.h"

MX_RECORD_FUNCTION_LIST mxd_network_doutput_record_function_list = {
	mxd_network_doutput_initialize_type,
	mxd_network_doutput_create_record_structures,
	mxd_network_doutput_finish_record_initialization,
	mxd_network_doutput_delete_record,
	NULL,
	mxd_network_doutput_read_parms_from_hardware,
	mxd_network_doutput_write_parms_to_hardware,
	mxd_network_doutput_open,
	mxd_network_doutput_close
};

MX_DIGITAL_OUTPUT_FUNCTION_LIST
			mxd_network_doutput_digital_output_function_list = {
	mxd_network_doutput_read,
	mxd_network_doutput_write
};

MX_RECORD_FIELD_DEFAULTS mxd_network_doutput_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_DIGITAL_OUTPUT_STANDARD_FIELDS,
	MXD_NETWORK_DOUTPUT_STANDARD_FIELDS
};

long mxd_network_doutput_num_record_fields
		= sizeof( mxd_network_doutput_record_field_defaults )
			/ sizeof( mxd_network_doutput_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_doutput_rfield_def_ptr
			= &mxd_network_doutput_record_field_defaults[0];

/* ===== */

MX_EXPORT mx_status_type
mxd_network_doutput_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_create_record_structures( MX_RECORD *record )
{
        const char fname[] = "mxd_network_doutput_create_record_structures()";

        MX_DIGITAL_OUTPUT *digital_output;
        MX_NETWORK_DOUTPUT *network_doutput;

        /* Allocate memory for the necessary structures. */

        digital_output = (MX_DIGITAL_OUTPUT *) malloc(sizeof(MX_DIGITAL_OUTPUT));

        if ( digital_output == (MX_DIGITAL_OUTPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_DIGITAL_OUTPUT structure." );
        }

        network_doutput = (MX_NETWORK_DOUTPUT *)
				malloc( sizeof(MX_NETWORK_DOUTPUT) );

        if ( network_doutput == (MX_NETWORK_DOUTPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_NETWORK_DOUTPUT structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = digital_output;
        record->record_type_struct = network_doutput;
        record->class_specific_function_list
			= &mxd_network_doutput_digital_output_function_list;

        digital_output->record = record;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_doutput_read( MX_DIGITAL_OUTPUT *doutput )
{
	const char fname[] = "mxd_network_doutput_read()";

	MX_NETWORK_DOUTPUT *network_doutput;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	unsigned long value;
	mx_status_type mx_status;

	network_doutput = (MX_NETWORK_DOUTPUT *)
				doutput->record->record_type_struct;

	if ( network_doutput == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_DOUTPUT pointer for digital output '%s' is NULL",
			doutput->record->name );
	}

	sprintf( rfname, "%s.value", network_doutput->remote_record_name );

	mx_status = mx_get_by_name( network_doutput->server_record,
				rfname, MXFT_HEX, &value );

	doutput->value = value;

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_doutput_write( MX_DIGITAL_OUTPUT *doutput )
{
	const char fname[] = "mxd_network_doutput_write()";

	MX_NETWORK_DOUTPUT *network_doutput;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	unsigned long value;
	mx_status_type mx_status;

	network_doutput = (MX_NETWORK_DOUTPUT *)
				doutput->record->record_type_struct;

	if ( network_doutput == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_DOUTPUT pointer for digital output '%s' is NULL",
			doutput->record->name );
	}

	value = doutput->value;

	sprintf( rfname, "%s.value", network_doutput->remote_record_name );

	mx_status = mx_put_by_name( network_doutput->server_record,
				rfname, MXFT_HEX, &value );

	return mx_status;
}

