/*
 * Name:    d_pmactc.c 
 *
 * Purpose: MX motor driver for controlling a Delta Tau PMAC motor
 *          through Tom Coleman's EPICS PMAC dual-ported RAM interface.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001, 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_epics.h"
#include "mx_motor.h"
#include "d_pmactc.h"

/* SBC-CAT specific data structures. */

MX_RECORD_FUNCTION_LIST mxd_pmac_tc_motor_record_function_list = {
	NULL,
	mxd_pmac_tc_motor_create_record_structures,
	mxd_pmac_tc_motor_finish_record_initialization,
	NULL,
	mxd_pmac_tc_motor_print_structure
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_tc_motor_motor_function_list = {
	mxd_pmac_tc_motor_motor_is_busy,
	mxd_pmac_tc_motor_move_absolute,
	mxd_pmac_tc_motor_get_position,
	NULL,
	mxd_pmac_tc_motor_soft_abort,
	mxd_pmac_tc_motor_immediate_abort,
	mxd_pmac_tc_motor_positive_limit_hit,
	mxd_pmac_tc_motor_negative_limit_hit,
	NULL,
	mxd_pmac_tc_motor_constant_velocity_move,
	mxd_pmac_tc_motor_get_parameter,
	mxd_pmac_tc_motor_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxd_pmac_tc_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_TC_MOTOR_STANDARD_FIELDS
};

long mxd_pmac_tc_motor_num_record_fields
		= sizeof( mxd_pmac_tc_motor_record_field_defaults )
			/ sizeof( mxd_pmac_tc_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_tc_motor_rfield_def_ptr
			= &mxd_pmac_tc_motor_record_field_defaults[0];

/* BioCAT specific data structures. */

MX_RECORD_FUNCTION_LIST mxd_pmac_bio_motor_record_function_list = {
	NULL,
	mxd_pmac_tc_motor_create_record_structures,
	mxd_pmac_tc_motor_finish_record_initialization,
	NULL,
	mxd_pmac_tc_motor_print_structure
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_bio_motor_motor_function_list = {
	mxd_pmac_tc_motor_motor_is_busy,
	mxd_pmac_tc_motor_move_absolute,
	mxd_pmac_tc_motor_get_position,
	NULL,
	mxd_pmac_tc_motor_soft_abort,
	mxd_pmac_tc_motor_immediate_abort,
	mxd_pmac_tc_motor_positive_limit_hit,
	mxd_pmac_tc_motor_negative_limit_hit,
	NULL,
	mxd_pmac_tc_motor_constant_velocity_move,
	mxd_pmac_bio_motor_get_parameter,
	mxd_pmac_bio_motor_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxd_pmac_bio_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_BIO_MOTOR_STANDARD_FIELDS
};

long mxd_pmac_bio_motor_num_record_fields
		= sizeof( mxd_pmac_bio_motor_record_field_defaults )
			/ sizeof( mxd_pmac_bio_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_bio_motor_rfield_def_ptr
			= &mxd_pmac_bio_motor_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_pmac_tc_motor_get_pointers( MX_MOTOR *motor,
			MX_PMAC_TC_MOTOR **pmac_tc_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_pmac_tc_motor_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac_tc_motor == (MX_PMAC_TC_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC_TC_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( motor->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MOTOR pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*pmac_tc_motor = (MX_PMAC_TC_MOTOR *) motor->record->record_type_struct;

	if ( *pmac_tc_motor == (MX_PMAC_TC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC_TC_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_tc_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_PMAC_TC_MOTOR *pmac_tc_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	pmac_tc_motor = (MX_PMAC_TC_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( pmac_tc_motor == (MX_PMAC_TC_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_PMAC_TC_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = pmac_tc_motor;

	if ( record->mx_type == MXT_MTR_PMAC_EPICS_TC ) {
		record->class_specific_function_list
				= &mxd_pmac_tc_motor_motor_function_list;
	} else {
		record->class_specific_function_list
				= &mxd_pmac_bio_motor_motor_function_list;
	}

	motor->record = record;

	/* A PMAC TC motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_finish_record_initialization( MX_RECORD *record )
{
	MX_MOTOR *motor;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor = (MX_MOTOR *) record->record_class_struct;

	motor->acceleration_type = MXF_MTR_ACCEL_TIME;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_tc_motor_print_structure()";

	MX_MOTOR *motor;
	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	double position, move_deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	pmac_tc_motor = (MX_PMAC_TC_MOTOR *) (record->record_type_struct);

	if ( pmac_tc_motor == (MX_PMAC_TC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_TC_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	switch( record->mx_type ) {
	case MXT_MTR_PMAC_EPICS_TC:
		fprintf(file,
		      "  Motor type                = PMAC_TC_MOTOR.\n\n");
		break;
	case MXT_MTR_PMAC_EPICS_BIO:
		fprintf(file,
		      "  Motor type                = PMAC_BIO_MOTOR.\n\n");
		break;
	default:
		fprintf(file,
		      "  Motor type                = %ld (this is a bug).\n\n",
						record->mx_type );
		break;
	}

	fprintf(file, "  name                      = %s\n", record->name);
	fprintf(file, "  actual position record    = %s\n",
			pmac_tc_motor->actual_position_record_name);
	fprintf(file, "  requested position record = %s\n",
			pmac_tc_motor->requested_position_record_name);
	fprintf(file, "  in position record        = %s\n",
			pmac_tc_motor->in_position_record_name);
	fprintf(file, "  abort record              = %s\n",
			pmac_tc_motor->abort_record_name);

	if ( record->mx_type == MXT_MTR_PMAC_EPICS_BIO ) {
		fprintf(file,
		      "  speed record name prefix  = %s\n",
			pmac_tc_motor->speed_record_name_prefix);
	}

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf( file, "  position                  = %g %s  (%g).\n",
		motor->position, motor->units,
		motor->raw_position.analog );
	fprintf( file, "  scale                     = %g %s per EPICS EGU.\n",
		motor->scale, motor->units );
	fprintf( file, "  offset                    = %g %s.\n",
		motor->offset, motor->units);
	fprintf( file, "  backlash                  = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf( file, "  negative limit            = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf( file, "  positive limit            = %g %s  (%g).\n",
	        motor->positive_limit, motor->units,
	        motor->raw_positive_limit.analog );

	move_deadband = motor->scale * motor->raw_move_deadband.analog;

	fprintf( file, "  move deadband             = %g %s  (%g).\n\n",
		move_deadband, motor->units,
		motor->raw_move_deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_motor_is_busy()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char pvname[60];
	short in_position_value;
	int i, max_retries;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor->busy = FALSE;

	strcpy( pvname, pmac_tc_motor->in_position_record_name );

	max_retries = 9;

	for ( i = 0; i < max_retries; i++ ) {

		MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_SHORT, 1, &in_position_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: read %hd from '%s'.",
				fname, in_position_value, pvname));

		switch ( in_position_value ) {
		case 0:
			motor->busy = TRUE;

			return MX_SUCCESSFUL_RESULT;

			break;
		case 1:
			/* Go back and check again. */

			break;
		default:
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"In position record '%s' for PMAC TC motor '%s' "
			"had an unexpected value of %d",
			pvname, motor->record->name, in_position_value );

			break;
		}
	}

	motor->busy = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_move_absolute()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	double new_destination;
	char pvname[60];
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	new_destination = motor->raw_destination.analog;

	/* Send the move command. */

	strcpy( pvname, pmac_tc_motor->requested_position_record_name );

	MX_DEBUG( 2,("%s: sending %g to '%s'.",
				fname, new_destination, pvname));

	mx_status = mx_caput_by_name( pvname,
					MX_CA_DOUBLE, 1, &new_destination );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_get_position()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	double scale, offset, raw_position;
	char pvname[60];
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	scale = motor->scale;
	offset = motor->offset;

	strcpy( pvname, pmac_tc_motor->actual_position_record_name );

	MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

	mx_status = mx_caget_by_name( pvname,
					MX_CA_DOUBLE, 1, &raw_position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor->raw_position.analog = raw_position;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_soft_abort()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char pvname[60];
	short stop_field;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strcpy( pvname, pmac_tc_motor->abort_record_name );

	/* Stop the motor move. */

	stop_field = 1;		/* Stop */

	MX_DEBUG( 2,("%s: sending %hd to '%s'.", fname, stop_field, pvname));

	mx_status = mx_caput_by_name( pvname, MX_CA_SHORT, 1, &stop_field );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_immediate_abort( MX_MOTOR *motor )
{
	return mxd_pmac_tc_motor_soft_abort( motor );
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_positive_limit_hit( MX_MOTOR *motor )
{
	motor->positive_limit_hit = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_negative_limit_hit( MX_MOTOR *motor )
{
	motor->negative_limit_hit = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_constant_velocity_move( MX_MOTOR *motor )
{
	mx_status_type mx_status;

	if ( motor->constant_velocity_move >= 0 ) {
		motor->raw_destination.analog = 10000;
	} else {
		motor->raw_destination.analog = -10000;
	}

	mx_status = mxd_pmac_tc_motor_move_absolute( motor );

	return mx_status;
}

/* Get and set parameter for SBC-CAT's version of the code. */

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_get_parameter()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char strcmd_pvname[60], strrsp_pvname[60];
	char command[41], response[41];
	int num_items;
	unsigned long pmac_speed;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s invoked for motor '%s' (type %ld) for parameter type '%s' (%d).",
		fname, motor->record->name, motor->record->mx_type,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	/* Send the parameter request command to the StrCmd EPICS variable. */

	sprintf( strcmd_pvname, "%s:StrCmd.VAL", pmac_tc_motor->pmac_name );

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( command, "I%d16", pmac_tc_motor->motor_number );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Parameter type %d is not supported by the driver for motor '%s'",
			motor->parameter_type, motor->record->name );
		break;
	}

	MX_DEBUG( 2,("%s: sending command '%s' to '%s'",
		fname, command, strcmd_pvname));

	mx_status = mx_caput_by_name( strcmd_pvname,
					MX_CA_STRING, 1, command );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Now read back the result from the StrRsp EPICS variable. */

	sprintf( strrsp_pvname, "%s:StrRsp.VAL", pmac_tc_motor->pmac_name );

	mx_status = mx_caget_by_name( strrsp_pvname,
					MX_CA_STRING, 1, response );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: received response '%s' from '%s'.",
		fname, response, strrsp_pvname));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		num_items = sscanf( response, "%lu", &pmac_speed );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"Unable to parse PMAC response '%s' for motor '%s'.",
				response, motor->record->name );
		}

		motor->raw_speed = (double) pmac_speed;
		break;
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_set_parameter()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char strcmd_pvname[60];
	char command[41];
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s invoked for motor '%s' (type %ld) for parameter type '%s' (%d).",
		fname, motor->record->name, motor->record->mx_type,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	/* Send the parameter request set to the StrCmd EPICS variable. */

	sprintf( strcmd_pvname, "%s:StrCmd.VAL", pmac_tc_motor->pmac_name );

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( command, "I%d16=%ld", pmac_tc_motor->motor_number,
						mx_round( motor->raw_speed ) );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Parameter type %d is not supported by the driver for motor '%s'",
			motor->parameter_type, motor->record->name );
		break;
	}

	MX_DEBUG( 2,("%s: sending command '%s' to '%s'",
		fname, command, strcmd_pvname));

	mx_status = mx_caput_by_name( strcmd_pvname, MX_CA_STRING, 1, command );

	MX_DEBUG( 2,("%s complete.", fname));

	return mx_status;
}

/* Get and set parameter for BioCAT's version of the code. */

MX_EXPORT mx_status_type
mxd_pmac_bio_motor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_bio_motor_get_parameter()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char pvname[60];
	double double_value;
	long long_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		if ( strlen(pmac_tc_motor->speed_record_name_prefix) <= 0 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Reading the motor speed is not configured for PMAC motor '%s'.",
				motor->record->name );
		}

		/* Tell EPICS to update the current value of the speed
		 * in its database.
		 */

		sprintf( pvname, "%s:Ix20:FAN.PROC",
				pmac_tc_motor->speed_record_name_prefix );

		long_value = 1;

		MX_DEBUG( 2,("%s: sending %ld to '%s'.",
				fname, long_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_LONG, 1, &long_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Now read the speed. */

		sprintf( pvname, "%s:Ix16:AI",
				pmac_tc_motor->speed_record_name_prefix );

		MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_DOUBLE, 1, &double_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: read %g from '%s'.",
				fname, double_value, pvname));

		motor->raw_speed = mx_divide_safely(
				pmac_tc_motor->speed_scale * double_value,
				  MX_PMAC_TC_MOTOR_SECONDS_PER_MILLISECOND );

		break;

	case MXLV_MTR_BASE_SPEED:
		motor->raw_base_speed = 0.0;
		break;

	case MXLV_MTR_RAW_ACCELERATION_PARAMETERS:

		/* Tell EPICS to update the current value of the acceleration
		 * time in its database.
		 */

		sprintf( pvname, "%s:Ix20:FAN.PROC",
				pmac_tc_motor->speed_record_name_prefix );

		long_value = 1;

		MX_DEBUG( 2,("%s: sending %ld to '%s'.",
				fname, long_value, pvname));

		mx_status = mx_caput_by_name( pvname,
						MX_CA_LONG, 1, &long_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Now read the acceleration time parameter. */

		sprintf( pvname, "%s:Ix20:LI",
				pmac_tc_motor->speed_record_name_prefix );

		MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_DOUBLE, 1, &double_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: read %g from '%s'.",
				fname, double_value, pvname));

		motor->raw_acceleration_parameters[0] = double_value
				* MX_PMAC_TC_MOTOR_SECONDS_PER_MILLISECOND;

		motor->raw_acceleration_parameters[1] = 0.0;
		motor->raw_acceleration_parameters[2] = 0.0;
		motor->raw_acceleration_parameters[3] = 0.0;
		break;

	case MXLV_MTR_ACCELERATION_TIME:
		mx_status = mx_motor_get_parameter( motor->record,
					MXLV_MTR_RAW_ACCELERATION_PARAMETERS );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		motor->acceleration_time =
				motor->raw_acceleration_parameters[0];
		break;

	default:
		return mx_motor_default_get_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_bio_motor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_bio_motor_set_parameter()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char pvname[60];
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		if ( strlen(pmac_tc_motor->speed_record_name_prefix) <= 0 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Setting the motor speed is not configured for PMAC motor '%s'.",
				motor->record->name );
		}
		sprintf( pvname, "%s:Ix16:AO",
				pmac_tc_motor->speed_record_name_prefix );

		double_value = mx_divide_safely(
		  motor->raw_speed * MX_PMAC_TC_MOTOR_SECONDS_PER_MILLISECOND,
				pmac_tc_motor->speed_scale );

		MX_DEBUG( 2,("%s: sending %g to '%s'.",
				fname, double_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_DOUBLE, 1, &double_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		break;

	case MXLV_MTR_BASE_SPEED:
		/* For BioCAT motors, setting the base speed is ignored. */

		break;

	case MXLV_MTR_SAVE_SPEED:
		mx_status = mx_motor_get_speed( motor->record, &double_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		motor->raw_saved_speed = motor->raw_speed;
		break;

	default:
		return mx_motor_default_set_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

