/*
 * Name:    i_win32_com.c
 *
 * Purpose: MX driver for Microsoft Win32 COM RS-232 devices.
 *
 * Author:  William Lavender
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 1999-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include "mx_unistd.h"

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rs232.h"
#include "mx_record.h"

#ifdef OS_WIN32	/* Include this only on Microsoft Win32 systems. */

/* The header file i_win32_com.h includes <windows.h> for us, so we don't have
 * to do that ourselves.
 */

#include "i_win32_com.h"

/* Do we show chars using MX_DEBUG() in the getchar and putchar functions? */

#define MXI_WIN32COM_DEBUG	FALSE

/* Define private functions to handle Microsoft Win32 COM ports. */

static mx_status_type win32com_write_parms( MX_RS232 *rs232 );

MX_RECORD_FUNCTION_LIST mxi_win32com_record_function_list = {
	mxi_win32com_initialize_type,
	mxi_win32com_create_record_structures,
	mxi_win32com_finish_record_initialization,
	mxi_win32com_delete_record,
	NULL,
	mxi_win32com_read_parms_from_hardware,
	mxi_win32com_write_parms_to_hardware,
	mxi_win32com_open,
	mxi_win32com_close
};

MX_RS232_FUNCTION_LIST mxi_win32com_rs232_function_list = {
	mxi_win32com_getchar,
	mxi_win32com_putchar,
	mxi_win32com_read,
	mxi_win32com_write,
	NULL,
	NULL,
	mxi_win32com_input_is_available,
	mxi_win32com_discard_unread_input,
	mxi_win32com_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_win32com_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_RS232_STANDARD_FIELDS,
	MXI_WIN32COM_STANDARD_FIELDS
};

long mxi_win32com_num_record_fields
		= sizeof( mxi_win32com_record_field_defaults )
			/ sizeof( mxi_win32com_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_win32com_rfield_def_ptr
			= &mxi_win32com_record_field_defaults[0];

/* ---- */

/* A private function for the use of the driver. */

static mx_status_type
mxi_win32com_get_pointers( MX_RS232 *rs232,
			MX_WIN32COM **win32com,
			const char *calling_fname )
{
	const char fname[] = "mxi_win32com_get_pointers()";

	MX_RECORD *win32com_record;

	if ( rs232 == (MX_RS232 *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_RS232 pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( win32com == (MX_WIN32COM **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_WIN32COM pointer passed by '%s' was NULL.",
			calling_fname );
	}

	win32com_record = rs232->record;

	if ( win32com_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The win32com_record pointer for the "
			"MX_RS232 pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*win32com = (MX_WIN32COM *) win32com_record->record_type_struct;

	if ( *win32com == (MX_WIN32COM *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_WIN32COM pointer for record '%s' is NULL.",
			win32com_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_win32com_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_win32com_create_record_structures()";

	MX_RS232 *rs232;
	MX_WIN32COM *win32com;

	/* Allocate memory for the necessary structures. */

	rs232 = (MX_RS232 *) malloc( sizeof(MX_RS232) );

	if ( rs232 == (MX_RS232 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_RS232 structure." );
	}

	win32com = (MX_WIN32COM *) malloc( sizeof(MX_WIN32COM) );

	if ( win32com == (MX_WIN32COM *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_WIN32COM structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = rs232;
	record->record_type_struct = win32com;
	record->class_specific_function_list
				= &mxi_win32com_rs232_function_list;

	rs232->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	/* Check to see if the RS-232 parameters are valid. */

	status = mx_rs232_check_port_parameters( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Mark the WIN32COM device as being closed. */

	record->flags |= ( ~MXF_REC_OPEN );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_write_parms_to_hardware( MX_RECORD *record )
{
	const char fname[] = "mxi_win32com_write_parms_to_hardware()";

	MX_RS232 *rs232;
	MX_WIN32COM *win32com;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	rs232 = (MX_RS232 *) (record->record_class_struct);

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = win32com_write_parms( rs232 );

	return status;
}

MX_EXPORT mx_status_type
mxi_win32com_open( MX_RECORD *record )
{
	const char fname[] = "mxi_win32com_open()";

	MX_RS232 *rs232;
	MX_WIN32COM *win32com;
	HANDLE com_port_handle;
	mx_status_type status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	rs232 = (MX_RS232 *) record->record_class_struct;

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Check the serial port name for validity. */

	if ( strnicmp( win32com->filename, "com", 3 ) != 0 ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The serial device name '%s' is illegal.", win32com->filename );
	}

	/* Initialize the WIN32COM driver on this port. */

	com_port_handle = CreateFile( win32com->filename,
		GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, 0 );

	if ( com_port_handle == INVALID_HANDLE_VALUE ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Error opening WIN32COM port '%s'.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	win32com->handle = com_port_handle;

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_close( MX_RECORD *record )
{
	const char fname[] = "mxi_win32com_close()";

	MX_RS232 *rs232;
	MX_WIN32COM *win32com;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	rs232 = (MX_RS232 *) (record->record_class_struct);

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Close the WIN32COM port. */

	CloseHandle( win32com->handle );

	/* Mark the port as closed. */

	win32com->handle = INVALID_HANDLE_VALUE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_getchar( MX_RS232 *rs232, char *c, int flags )
{
	const char fname[] = "mxi_win32com_getchar()";

	MX_WIN32COM *win32com;
	BOOL win32_status;
	DWORD num_chars;
	int input_available;
	mx_status_type status;

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( flags & MXF_232_NOWAIT ) {

		status = mxi_win32com_input_is_available(
					rs232, &input_available );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( input_available == FALSE ) {
			return mx_error_quiet( MXE_NOT_READY, fname,
				"Failed to read a character from port '%s'.",
				rs232->record->name );
		}
	}

	win32_status = ReadFile( win32com->handle, c, 1, &num_chars, 0 );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Error while trying to read a character from '%s'.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

#if MXI_WIN32COM_DEBUG
	MX_DEBUG(-2,("%s: received c = 0x%x, '%c'", fname, *c, *c));
#endif

	/* mxi_win32com_getchar() is often used to test whether there is
	 * input ready on the COM port.  Normally, it is not desireable
	 * to broadcast a message to the world when this fails, so we
	 * use mx_error_quiet() rather than mx_error().
	 */

	if ( num_chars != 1 ) {
		return mx_error_quiet( MXE_NOT_READY, fname,
			"Failed to read a character from port '%s'.",
			rs232->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_putchar( MX_RS232 *rs232, char c, int flags )
{
	const char fname[] = "mxi_win32com_putchar()";

	MX_WIN32COM *win32com;
	BOOL win32_status;
	DWORD num_chars;
	mx_status_type status;

#if MXI_WIN32COM_DEBUG
	MX_DEBUG(-2,("%s: sending c = 0x%x, '%c'", fname, c, c));
#endif

	if ( flags & MXF_232_NOWAIT ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		  "Non-blocking Win32 COM port I/O not yet implemented.");
	}

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	win32_status = WriteFile( win32com->handle, &c, 1, &num_chars, 0 );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Attempt to send a character to port '%s' failed.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	/* Force the character to be transmitted to the device. */

	win32_status = FlushFileBuffers( win32com->handle );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Attempt to flush the output for port '%s' failed.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_read( MX_RS232 *rs232,
			char *buffer,
			size_t max_bytes_to_read,
			size_t *bytes_read )
{
	const char fname[] = "mxi_win32com_read()";

	mx_status_type status;
	long i;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* This is slow, but simple.  It currently waits forever until
	 * 'max_chars' are read.  Think about doing something better later.
	 */

	for ( i = 0; i < max_bytes_to_read; i++ ) {
		status = mxi_win32com_getchar(rs232, &buffer[i], MXF_232_WAIT);

		if ( status.code != MXE_SUCCESS ) {
			i--;
			continue;
		}
	}

	*bytes_read = max_bytes_to_read;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_write( MX_RS232 *rs232,
			char *buffer,
			size_t max_bytes_to_write,
			size_t *bytes_written )
{
	const char fname[] = "mxi_win32com_write()";

	mx_status_type status;
	long i;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* This is slow, but simple. */

	for ( i = 0; i < max_bytes_to_write; i++ ) {
		status = mxi_win32com_putchar( rs232, buffer[i], MXF_232_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			return status;
		}
	}

	*bytes_written = max_bytes_to_write;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_input_is_available( MX_RS232 *rs232, int *input_available )
{
	const char fname[] = "mxi_win32com_input_is_available()";

	MX_WIN32COM *win32com;
	DWORD error_mask;
	COMSTAT com_status;
	BOOL win32_status;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.",fname));

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	win32_status = ClearCommError( win32com->handle,
						&error_mask, &com_status );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Error while trying to check for input from '%s'.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	if ( com_status.cbInQue > 0 ) {

		*input_available = TRUE;
	} else {
		*input_available = FALSE;
	}

	MX_DEBUG( 2,("%s complete.",fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_discard_unread_input( MX_RS232 *rs232, int debug_flag )
{
	const char fname[] = "mxi_win32com_discard_unread_input()";

	MX_WIN32COM *win32com;
	BOOL win32_status;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.",fname));

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	win32_status = PurgeComm( win32com->handle,
					( PURGE_RXABORT | PURGE_RXCLEAR ) );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Error while trying to discard unread input for '%s'.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	MX_DEBUG( 2,("%s complete.",fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_win32com_discard_unwritten_output( MX_RS232 *rs232, int debug_flag )
{
	const char fname[] = "mxi_win32com_discard_unwritten_output()";

	MX_WIN32COM *win32com;
	BOOL win32_status;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.",fname));

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	win32_status = PurgeComm( win32com->handle,
					( PURGE_TXABORT | PURGE_TXCLEAR ) );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Error while trying to discard unwritten output for '%s'.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	MX_DEBUG( 2,("%s complete.",fname));

	return MX_SUCCESSFUL_RESULT;
}

/* ==== Microsoft Win32 I/O functions ==== */

static mx_status_type
win32com_write_parms( MX_RS232 *rs232 )
{
	const char fname[] = "win32com_write_parms()";

	MX_RECORD *record;
	MX_WIN32COM *win32com;
	BOOL win32_status;
	DCB dcb;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.",fname));

	status = mxi_win32com_get_pointers( rs232, &win32com, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( win32com->handle == INVALID_HANDLE_VALUE ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"WIN32COM port device '%s' is not open.",
		rs232->record->name );
	}

	/* Get the current settings of the COM port. */

	win32_status = GetCommState( win32com->handle, &dcb );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Attempt to get current port settings for '%s' failed.  "
		"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}

	/* Set the port speed. */

	switch( rs232->speed ) {
	case 256000: dcb.BaudRate = CBR_256000; break;
	case 128000: dcb.BaudRate = CBR_128000; break;
	case 115200: dcb.BaudRate = CBR_115200; break;
	case 57600:  dcb.BaudRate = CBR_57600;  break;
	case 56000:  dcb.BaudRate = CBR_56000;  break;
	case 38400:  dcb.BaudRate = CBR_38400;  break;
	case 19200:  dcb.BaudRate = CBR_19200;  break;
	case 14400:  dcb.BaudRate = CBR_14400;  break;
	case 9600:   dcb.BaudRate = CBR_9600;   break;
	case 4800:   dcb.BaudRate = CBR_4800;   break;
	case 2400:   dcb.BaudRate = CBR_2400;   break;
	case 1200:   dcb.BaudRate = CBR_1200;   break;
	case 600:    dcb.BaudRate = CBR_600;    break;
	case 300:    dcb.BaudRate = CBR_300;    break;
	case 110:    dcb.BaudRate = CBR_110;    break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported WIN32COM speed = %d for WIN32COM '%s'.", 
		rs232->speed, rs232->record->name);

		break;
	}

	/* Set the word size. */

	if ( rs232->word_size >= 4 && rs232->word_size <= 8 ) {
		dcb.ByteSize = rs232->word_size;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported WIN32COM word size = %d for WIN32COM '%s'.", 
		rs232->word_size, rs232->record->name);
	}

	/* Set the parity. */

	dcb.fParity = TRUE;

	switch( rs232->parity ) {
	case MXF_232_EVEN_PARITY:  dcb.Parity = EVENPARITY;  break;
	case MXF_232_ODD_PARITY:   dcb.Parity = ODDPARITY;   break;
	case MXF_232_MARK_PARITY:  dcb.Parity = MARKPARITY;  break;
	case MXF_232_SPACE_PARITY: dcb.Parity = SPACEPARITY; break;

	case MXF_232_NO_PARITY:
		dcb.Parity = NOPARITY;
		dcb.fParity = FALSE;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported WIN32COM parity = '%c' for WIN32COM '%s'.", 
		rs232->parity, rs232->record->name);

		break;
	}

	/* Set the stop bits. */

	switch( rs232->stop_bits ) {
	case 1:  dcb.StopBits = ONESTOPBIT; break;
	case 2:  dcb.StopBits = TWOSTOPBITS; break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported WIN32COM stop bits = %d for WIN32COM '%s'.", 
		rs232->stop_bits, rs232->record->name);

		break;
	}

	/* Set the flow control parameters. */

	dcb.fInX = FALSE;
	dcb.fOutX = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;

	switch( rs232->flow_control ) {
	case MXF_232_NO_FLOW_CONTROL:
		break;

	case MXF_232_SOFTWARE_FLOW_CONTROL:
		dcb.fInX = TRUE;
		dcb.fOutX = TRUE;
		dcb.XonLim = 50;
		dcb.XoffLim = 200;
		dcb.XonChar = 17;
		dcb.XoffChar = 19;
		break;

	case MXF_232_RTS_CTS_FLOW_CONTROL:
	case MXF_232_HARDWARE_FLOW_CONTROL:
		dcb.fOutxCtsFlow = TRUE;
		dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
		break;

	case MXF_232_DTR_DSR_FLOW_CONTROL:
		dcb.fOutxDsrFlow = TRUE;
		dcb.fDtrControl = DTR_CONTROL_HANDSHAKE;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported WIN32COM flow control = %c for WIN32COM '%s'.", 
		rs232->flow_control, rs232->record->name);

		break;
	}

	/* Everything is set, so change the configuration of the COM port.*/

	win32_status = SetCommState( win32com->handle, &dcb );

	if ( win32_status == 0 ) {
		DWORD last_error_code;
		TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

		last_error_code = GetLastError();

		mx_win32_error_message( last_error_code,
			message_buffer, sizeof(message_buffer) );

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Attempt to set current port settings for '%s' failed.  "
			"Win32 error code = %ld, error message = '%s'",
			win32com->filename, last_error_code, message_buffer );
	}
		
	MX_DEBUG( 2,("%s complete.",fname));

	return MX_SUCCESSFUL_RESULT;
}

#endif /* OS_WIN32 */
