/*
 * Name:    mx_analog_input.c
 *
 * Purpose: MX function library for analog input devices.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_analog_input.h"

static mx_status_type
mx_analog_input_get_pointers( MX_RECORD *record,
				MX_ANALOG_INPUT **analog_input,
				MX_ANALOG_INPUT_FUNCTION_LIST **flist_ptr,
				const char *calling_fname )
{
	const char fname[] = "mx_analog_input_get_pointers()";

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( record->mx_class != MXC_ANALOG_INPUT ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not an analog input record.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			record->name, calling_fname,
			record->mx_superclass,
			record->mx_class,
			record->mx_type );
	}

	if ( analog_input != (MX_ANALOG_INPUT **) NULL ) {
		*analog_input = (MX_ANALOG_INPUT *)
						(record->record_class_struct);

		if ( *analog_input == (MX_ANALOG_INPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_ANALOG_INPUT pointer for record '%s' passed by '%s' is NULL.",
				record->name, calling_fname );
		}
	}

	if ( flist_ptr != (MX_ANALOG_INPUT_FUNCTION_LIST **) NULL ) {
		*flist_ptr = (MX_ANALOG_INPUT_FUNCTION_LIST *)
				(record->class_specific_function_list);

		if ( *flist_ptr == (MX_ANALOG_INPUT_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_ANALOG_INPUT_FUNCTION_LIST ptr for record '%s' passed by '%s' is NULL.",
				record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_analog_input_read( MX_RECORD *record, double *value )
{
	const char fname[] = "mx_analog_input_read()";

	MX_ANALOG_INPUT *analog_input;
	MX_ANALOG_INPUT_FUNCTION_LIST *function_list;
	double raw_value;
	mx_status_type ( *read_fn ) ( MX_ANALOG_INPUT * );
	mx_status_type status;

	status = mx_analog_input_get_pointers( record, &analog_input,
						&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_fn = function_list->read;

	if ( read_fn == NULL ){
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_ANALOG_INPUT ptr 0x%p is NULL.",
			analog_input);
	}

	status = (*read_fn)( analog_input );

	if ( analog_input->subclass == MXT_AIN_LONG ) {
		raw_value = (double) analog_input->raw_value.long_value;
	} else {
		raw_value = analog_input->raw_value.double_value;
	}

	analog_input->value = analog_input->offset
			    + analog_input->scale * raw_value;

	if ( value != NULL ) {
		*value = analog_input->value;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_analog_input_read_raw_long( MX_RECORD *record, long *raw_value )
{
	const char fname[] = "mx_analog_input_read_raw_long()";

	MX_ANALOG_INPUT *analog_input;
	MX_ANALOG_INPUT_FUNCTION_LIST *function_list;
	long local_raw_value;
	mx_status_type ( *read_fn ) ( MX_ANALOG_INPUT * );
	mx_status_type status;

	status = mx_analog_input_get_pointers( record, &analog_input,
						&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( analog_input->subclass != MXT_AIN_LONG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"This function may only be used with analog inputs that return values "
	"as long integers.  The record '%s' returns values as doubles so "
	"you should use mx_analog_input_read_raw_double() with this record.",
			record->name );
	}

	read_fn = function_list->read;

	if ( read_fn == NULL ){
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_ANALOG_INPUT ptr 0x%p is NULL.",
			analog_input);
	}

	status = (*read_fn)( analog_input );

	local_raw_value = analog_input->raw_value.long_value;

	analog_input->value = analog_input->offset + analog_input->scale
					* (double) local_raw_value;

	if ( raw_value != NULL ) {
		*raw_value = local_raw_value;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_analog_input_read_raw_double( MX_RECORD *record, double *raw_value )
{
	const char fname[] = "mx_analog_input_read_raw_double()";

	MX_ANALOG_INPUT *analog_input;
	MX_ANALOG_INPUT_FUNCTION_LIST *function_list;
	double local_raw_value;
	mx_status_type ( *read_fn ) ( MX_ANALOG_INPUT * );
	mx_status_type status;

	status = mx_analog_input_get_pointers( record, &analog_input,
						&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( analog_input->subclass != MXT_AIN_DOUBLE ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"This function may only be used with analog inputs that return values "
	"as doubles.  The record '%s' returns values as long integers so "
	"you should use mx_analog_input_read_raw_long() with this record.",
			record->name );
	}

	read_fn = function_list->read;

	if ( read_fn == NULL ){
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_ANALOG_INPUT ptr 0x%p is NULL.",
			analog_input);
	}

	status = (*read_fn)( analog_input );

	local_raw_value = analog_input->raw_value.double_value;

	analog_input->value = analog_input->offset + analog_input->scale
					* local_raw_value;

	if ( raw_value != NULL ) {
		*raw_value = local_raw_value;
	}

	return status;
}

