/*
 * Name:     mx_net.h
 *
 * Purpose:  Header file for the MX network protocol.
 *
 * Author:   William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2000, 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __MX_NET_H__
#define __MX_NET_H__

#include "mx_types.h"

/* Current size of an MX network header. */

#define MX_NETWORK_HEADER_LENGTH_VALUE	( 5 * sizeof(mx_uint32_type) )

/*
 * Maximum size of an MX network message.  This limit needs to be eliminated,
 * but for now we are stuck with it.
 */

#define MX_NETWORK_MAXIMUM_MESSAGE_SIZE  163840L

/*
 * Define the data type that contains MX network messages.
 *
 * The following union is defined so that the address of the first byte
 * will be correctly aligned on a 32-bit word boundary for those computer 
 * architectures that require such things.
 */

typedef union {
	mx_uint32_type header[MX_NETWORK_HEADER_LENGTH_VALUE];
	char buffer[MX_NETWORK_MAXIMUM_MESSAGE_SIZE];
} MX_NETWORK_MESSAGE_BUFFER;

/*
 * MX network server data structure.
 */

typedef struct {
	MX_RECORD *record;
	MX_NETWORK_MESSAGE_BUFFER *message_buffer;

	unsigned long remote_mx_version;
} MX_NETWORK_SERVER;

typedef struct {
	mx_status_type ( *receive_message ) ( MX_NETWORK_SERVER *server,
						unsigned long buffer_length,
						void *buffer );
	mx_status_type ( *send_message ) ( MX_NETWORK_SERVER *server,
						void *buffer );
	mx_status_type ( *reconnect_if_down ) ( MX_NETWORK_SERVER *server );
} MX_NETWORK_SERVER_FUNCTION_LIST;

/* Definition of the fields in a header. */

#define MX_NETWORK_MAGIC		0
#define MX_NETWORK_HEADER_LENGTH	1
#define MX_NETWORK_MESSAGE_LENGTH	2
#define MX_NETWORK_MESSAGE_TYPE		3
#define MX_NETWORK_STATUS_CODE		4

/* Definition of network message type flags. */

#define MX_NETMSG_ERROR_FLAG		0x8000000
#define MX_NETMSG_SERVER_RESPONSE_FLAG	0x4000000
#define MX_NETMSG_SERVER_CALLBACK_FLAG	0x2000000

#define mx_server_response(x)	((x) | MX_NETMSG_SERVER_RESPONSE_FLAG)
#define mx_server_callback(x)	((x) | MX_NETMSG_SERVER_CALLBACK_FLAG)

/* Definition of network message types. */

#define MX_NETMSG_UNEXPECTED_ERROR	0x1

/*
 * Messages from clients to servers.
 *
 * The response from the server will have the same value with the
 * MX_NETMSG_SERVER_RESPONSE_FLAG bit set.
 *
 */

#define MX_NETMSG_GET_ARRAY_BY_NAME	0x1001
#define MX_NETMSG_PUT_ARRAY_BY_NAME	0x1002

#define MX_NETMSG_GET_FIELD_TYPE	0x2005

#define MX_NETMSG_SET_CLIENT_INFO	0x3001

/* The MX_NETWORK_MAGIC header field should always contain the
 * MX_NETWORK_MAGIC_VALUE.  If it does not, there has been
 * a protocol error somewhere.
 */

#define MX_NETWORK_MAGIC_VALUE		0xe7e9fcfe

MX_API mx_status_type mx_network_receive_message( MX_RECORD *server_record,
				unsigned long buffer_length, void *buffer );

MX_API mx_status_type mx_network_send_message( MX_RECORD *server_record,
					void *buffer );

MX_API mx_status_type mx_network_reconnect_if_down( MX_RECORD *server_record );

MX_API void mx_network_display_message_buffer( void *buffer );

#define mx_get_by_name( s, r, d, v ) \
		mx_get_array_ascii( (s), (r), (d), 0, NULL, (v) )

#define mx_put_by_name( s, r, d, v ) \
		mx_put_array_ascii( (s), (r), (d), 0, NULL, (v) )

#define mx_get_array_by_name	mx_get_array_ascii

#define mx_put_array_by_name	mx_put_array_ascii

MX_API mx_status_type mx_get_array_ascii( MX_RECORD *server_record,
			char *remote_record_field_name,
			long datatype, 
			long num_dimensions,
			long *dimension,
			void *value );

MX_API mx_status_type mx_put_array_ascii( MX_RECORD *server_record,
			char *remote_record_field_name,
			long datatype, 
			long num_dimensions,
			long *dimension,
			void *value );

MX_API mx_status_type mx_get_field_array_ascii( MX_RECORD *server_record,
			char *remote_record_field_name,
			MX_RECORD_FIELD *local_field,
			void *value_ptr );

MX_API mx_status_type mx_put_field_array_ascii( MX_RECORD *server_record,
			char *remote_record_field_name,
			MX_RECORD_FIELD *local_field,
			void *value_ptr );

MX_API mx_status_type mx_get_field_type( MX_RECORD *server_record,
			char *remote_record_field_name,
			long max_dimensions,
			long *datatype,
			long *num_dimensions,
			long *dimension_array );

MX_API mx_status_type mx_set_client_info( MX_RECORD *server_record,
			char *username,
			char *program_name );

#endif /* __MX_NET_H__ */
