/*
 * Name:     mx_plot.c
 *
 * Purpose:  Support for plotting a data file during a scan.
 *
 * Author:   William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "mx_record.h"
#include "mx_plot.h"
#include "p_none.h"
#include "p_child.h"
#include "p_gnuplot.h"

#include "mx_motor.h"
#include "p_gnuplot_xafs.h"

MX_PLOT_TYPE_ENTRY mx_plot_type_list[] = {
	{ MXP_NONE, "none", &mxp_none_function_list },
	{ MXP_CHILD, "child", &mxp_child_function_list },
	{ MXP_GNUPLOT, "gnuplot", &mxp_gnuplot_function_list },
	{ MXP_GNUXAFS, "gnuxafs", &mxp_gnuxafs_function_list },
	{ -1, "", NULL }
};

MX_EXPORT mx_status_type mx_get_plot_type_by_name( 
	MX_PLOT_TYPE_ENTRY *plot_type_list,
	char *name,
	MX_PLOT_TYPE_ENTRY **plot_type_entry )
{
	const char fname[] = "mx_get_plot_type_by_name()";

	char *list_name;
	int i;

	if ( plot_type_list == NULL ) {
		*plot_type_entry = NULL;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Plot type list passed was NULL." );
	}

	if ( name == NULL ) {
		*plot_type_entry = NULL;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Plot name pointer passed is NULL.");
	}

	for ( i=0; ; i++ ) {
		/* Check for the end of the list. */

		if ( plot_type_list[i].type < 0 ) {
			*plot_type_entry = NULL;

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Plot type '%s' was not found.", name );
		}

		list_name = plot_type_list[i].name;

		if ( list_name == NULL ) {
			*plot_type_entry = NULL;

			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"NULL name ptr for plot type %ld.",
				plot_type_list[i].type );
		}

		if ( strcmp( name, list_name ) == 0 ) {
			*plot_type_entry = &( plot_type_list[i] );

			MX_DEBUG( 8,
			("mx_get_plot_type_by_name: ptr = 0x%p, type = %ld",
				*plot_type_entry, (*plot_type_entry)->type));

			return MX_SUCCESSFUL_RESULT;
		}
	}
}

MX_EXPORT mx_status_type mx_get_plot_type_by_value( 
	MX_PLOT_TYPE_ENTRY *plot_type_list,
	long plot_type,
	MX_PLOT_TYPE_ENTRY **plot_type_entry )
{
	const char fname[] = "mx_get_plot_type_by_value()";

	int i;

	if ( plot_type_list == NULL ) {
		*plot_type_entry = NULL;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Plot type list passed was NULL." );
	}

	if ( plot_type <= 0 ) {
		*plot_type_entry = NULL;

		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Plot type %ld is illegal.", plot_type );
	}

	for ( i=0; ; i++ ) {
		/* Check for the end of the list. */

		if ( plot_type_list[i].type < 0 ) {
			*plot_type_entry = NULL;

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Plot type %ld was not found.", plot_type );
		}

		if ( plot_type_list[i].type == plot_type ) {
			*plot_type_entry = &( plot_type_list[i] );

			MX_DEBUG( 8,
			("mx_get_plot_type_by_value: ptr = 0x%p, type = %ld",
				*plot_type_entry, (*plot_type_entry)->type));

			return MX_SUCCESSFUL_RESULT;
		}
	}
}

MX_EXPORT mx_status_type
mx_plot_open( MX_PLOT *plot )
{
	const char fname[] = "mx_plot_open()";

	MX_PLOT_TYPE_ENTRY *plot_type_entry;
	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT * );
	mx_status_type status;

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	status = mx_get_plot_type_by_value(
			mx_plot_type_list, plot->type, &plot_type_entry );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( plot_type_entry == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"mx_get_plot_type_by_value() returned a NULL plot_type_entry pointer." );
	}

	flist = plot_type_entry->plot_function_list;

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot type '%s' is NULL",
			plot_type_entry->name );
	}

	/* Save a copy of the function list pointer. */

	plot->plot_function_list = flist;

	plot->plot_type_struct = NULL;

	/* Now invoke the open function. */

	fptr = flist->open;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"open function pointer for plot type '%s' is NULL",
			plot_type_entry->name );
	}

	status = (*fptr) ( plot );

	return status;
}

MX_EXPORT mx_status_type
mx_plot_close( MX_PLOT *plot )
{
	const char fname[] = "mx_plot_close()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT * );
	mx_status_type status;

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the close function. */

	fptr = flist->close;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"close function pointer for plot type '%s' is NULL.",
			plot->typename );
	}

	status = (*fptr) ( plot );

	return status;
}

MX_EXPORT mx_status_type
mx_add_measurement_to_plot_buffer( MX_PLOT *plot )
{
	const char fname[] = "mx_add_measurement_to_plot_buffer()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT * );
	mx_status_type status;

	MX_DEBUG(8,("%s invoked.", fname));

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the add_measurement_to_plot_buffer function. */

	fptr = flist->add_measurement_to_plot_buffer;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"add_measurement_to_plot_buffer function pointer for plot is NULL." );
	}

	status = (*fptr) ( plot );

	return status;
}

MX_EXPORT mx_status_type
mx_add_array_to_plot_buffer( MX_PLOT *plot,
		long position_type, long num_positions, void *position_array,
		long data_type, long num_data_points, void *data_array )
{
	const char fname[] = "mx_add_array_to_plot_buffer()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT *, long, long, void *,
						long, long, void * );
	mx_status_type status;

	MX_DEBUG(8,("%s invoked.", fname));

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the add_array_to_plot_buffer function. */

	fptr = flist->add_array_to_plot_buffer;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"add_array_to_plot_buffer function pointer for plot is NULL." );
	}

	status = (*fptr) ( plot, position_type, num_positions, position_array,
				data_type, num_data_points, data_array );

	return status;
}

MX_EXPORT mx_status_type
mx_display_plot( MX_PLOT *plot )
{
	const char fname[] = "mx_display_plot()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT * );
	mx_status_type status;

	MX_DEBUG(8,("%s invoked.", fname));

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the display_plot function. */

	fptr = flist->display_plot;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"display_plot function pointer for plot is NULL." );
	}

	status = (*fptr) ( plot );

	return status;
}

MX_EXPORT mx_status_type
mx_plot_set_x_range( MX_PLOT *plot, double x_min, double x_max )
{
	const char fname[] = "mx_plot_set_x_range()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT *, double, double );
	mx_status_type status;

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the set_x_range function. */

	fptr = flist->set_x_range;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_x_range function pointer for plot is NULL." );
	}

	status = (*fptr) ( plot, x_min, x_max );

	return status;
}

MX_EXPORT mx_status_type
mx_plot_set_y_range( MX_PLOT *plot, double y_min, double y_max )
{
	const char fname[] = "mx_plot_set_y_range()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT *, double, double );
	mx_status_type status;

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the set_y_range function. */

	fptr = flist->set_y_range;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_y_range function pointer for plot is NULL." );
	}

	status = (*fptr) ( plot, y_min, y_max );

	return status;
}

MX_EXPORT mx_status_type
mx_plot_start_plot_section( MX_PLOT *plot )
{
	const char fname[] = "mx_plot_start_plot_section()";

	MX_PLOT_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_PLOT * );
	mx_status_type status;

	MX_DEBUG( 8,("%s invoked.",fname));

	if ( plot == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_PLOT pointer passed was NULL.");
	}

	flist = (MX_PLOT_FUNCTION_LIST *) (plot->plot_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PLOT_FUNCTION_LIST pointer for plot is NULL");
	}

	/* Now invoke the start_plot_section function. */

	fptr = flist->start_plot_section;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start_plot_section function pointer for plot is NULL" );
	}

	status = (*fptr) ( plot );

	return status;
}

MX_EXPORT int
mx_plotting_is_enabled( MX_RECORD *record )
{
	MX_RECORD *list_head_record;
	MX_LIST_HEAD *list_head_struct;
	int enabled;

	if ( record == (MX_RECORD *) NULL )
		return FALSE;

	list_head_record = (MX_RECORD *)(record->list_head);

	list_head_struct = (MX_LIST_HEAD *)
				(list_head_record->record_superclass_struct);

	enabled = list_head_struct->plotting_enabled;

	return enabled;
}

MX_EXPORT void
mx_set_plot_enable( MX_RECORD *record, int enabled )
{
	MX_RECORD *list_head_record;
	MX_LIST_HEAD *list_head_struct;

	if ( record == (MX_RECORD *) NULL )
		return;

	list_head_record = (MX_RECORD *)(record->list_head);

	list_head_struct = (MX_LIST_HEAD *)
				(list_head_record->record_superclass_struct);

	list_head_struct->plotting_enabled = enabled;

	return;
}

MX_EXPORT MX_PLOT_TYPE_ENTRY *
mx_plot_get_type_list( void )
{
	return mx_plot_type_list;
}

