/*
 * Name:    mx_udelay.c
 *
 * Purpose: In general, the mx_sleep(), mx_msleep(), and mx_usleep() family
 *          of functions perform their tasks by yielding control to the
 *          operating system and requesting that they be woken up again
 *          by the OS at a later time.  This usually means that you cannot
 *          create delays this way that are shorter than the minimum process
 *          scheduling interval for the operating system (100 Hz for most
 *          versions of Linux).
 *
 *          mx_udelay() instead attempts to implement microsecond delays
 *          that do _not_ explicitly return control to the operating system.
 *          This allows for delays that are much shorter than the scheduler
 *          would permit.  This is not possible on all operating systems,
 *          so mx_udelay() reverts back to using mx_usleep() in that case.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_osdef.h"
#include "mx_util.h"

#if defined(OS_LINUX) && defined(__i386__)

/******* x86 Linux *******/

/* The following was inspired by the Linux I/O port programming mini-HOWTO. */

#define MX_CPU_UNKNOWN		0
#define MX_CPU_USE_XCHG		1
#define MX_CPU_USE_RDTSC	2

static int mx_cpu_delay_type = MX_CPU_UNKNOWN;

static double mx_delay_instructions_per_microsecond = 0.0;

static __inline__ unsigned long long int
mx_rdtsc( void )
{
	unsigned long long int x;

	/* The following generates inline an RDTSC instruction
	 * and puts the results in the variable "x".
	 */

	__asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));

	return x;
}

static __inline__ void
mx_xchg_delay( void )
{
	/* The following takes three clock cycles on i386 and i486 machines.
	 * We will be using RDTSC on other x86 CPU types.
	 */

	__asm__ volatile ("xchg %bx,%bx");

	return;
}

MX_EXPORT void
mx_udelay( unsigned long microseconds )
{
	if ( mx_cpu_delay_type == MX_CPU_UNKNOWN )
		mx_udelay_init();

	if ( mx_cpu_delay_type == MX_CPU_USE_RDTSC ) {

		/* Warning: The following logic has the fundamental flaw
		 * that the TSC counter will overflow in around 584 years
		 * on a 1 GHz x86 machine, but I expect to be retired long
		 * before that.  (WML)
		 */

		unsigned long long int tsc_start_value, tsc_end_value;
		unsigned long long int tsc_delay_ticks;

		tsc_delay_ticks = (unsigned long long int)
	( mx_delay_instructions_per_microsecond * (double) microseconds );

		tsc_start_value = mx_rdtsc();

		tsc_end_value = tsc_start_value + tsc_delay_ticks;

		/* Sit in a loop until we reach the end of the delay time. */

		while ( mx_rdtsc() < tsc_end_value )
			;				/* Do nothing */

	} else {
		/* This code does not work as well as the RDTSC code since
		 * it does not take into account delays due to the process
		 * being preempted by the kernel.
		 */

		unsigned long long int i, num_times_to_loop;

		num_times_to_loop = (unsigned long long int)
	( mx_delay_instructions_per_microsecond * (double) microseconds );

		for ( i = 0; i < num_times_to_loop; i++ ) {
			mx_xchg_delay();
		}
	}
	return;
}

MX_EXPORT void
mx_udelay_init( void )
{
	static const char fname[] = "mx_udelay_init()";

	char buffer[500];
	char *ptr, *ptr2;
	int cpu_family, have_tsc, num_items;
	double cpu_mhz;

	/* Reading /proc/cpuinfo is the only vaguely portable way I know of
	 * for getting the necessary information.
	 */

	FILE *cpuinfo;

	cpuinfo = fopen( "/proc/cpuinfo", "r" );

	if ( cpuinfo == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
		"/proc/cpuinfo could not be found or is unreadable.  "
		"This may indicate that there is something seriously "
		"wrong with your Linux system." );

		return;
	}

	cpu_family = 0;
	have_tsc = FALSE;
	cpu_mhz = 0.0;

	fgets( buffer, sizeof buffer, cpuinfo );

	while ( !feof(cpuinfo) && !ferror(cpuinfo) ) {

		if ( strncmp( buffer, "cpu family", 10 ) == 0 ) {

			ptr = strchr( buffer, ':' );

			if ( ptr == NULL ) {
				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
				"Could not find a ':' character in a line "
				"read from /proc/cpuinfo.  line = '%s'",
					buffer );
			} else {
				ptr++;

				num_items = sscanf( ptr, "%d", &cpu_family );

				if ( num_items != 1 ) {
				    (void) mx_error( MXE_UNPARSEABLE_STRING,
					fname,
				"Could not find the cpu family number in a "
				"line read from /proc/cpuinfo.  line = '%s'",
					buffer );
				}
			}
		} else
		if ( strncmp( buffer, "cpu MHz", 7 ) == 0 ) {

			ptr = strchr( buffer, ':' );

			if ( ptr == NULL ) {
				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
				"Could not find a ':' character in a line "
				"read from /proc/cpuinfo.  line = '%s'",
					buffer );
			} else {
				ptr++;

				num_items = sscanf( ptr, "%lg", &cpu_mhz );

				if ( num_items != 1 ) {
				    (void) mx_error( MXE_UNPARSEABLE_STRING,
					fname,
				"Could not find the cpu MHz in a "
				"line read from /proc/cpuinfo.  line = '%s'",
					buffer );
				}
			}
		} else
		if ( strncmp( buffer, "flags", 5 ) == 0 ) {

			ptr = strchr( buffer, ':' );

			if ( ptr == NULL ) {
				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
				"Could not find a ':' character in a line "
				"read from /proc/cpuinfo.  line = '%s'",
					buffer );
			} else {
				ptr++;

				ptr2 = strstr( ptr, "tsc" );

				if ( ptr2 != NULL ) {
					have_tsc = TRUE;
				}
			}
		}

		fgets( buffer, sizeof buffer, cpuinfo );
	}

	fclose( cpuinfo );

	if ( have_tsc ) {
		mx_cpu_delay_type = MX_CPU_USE_RDTSC;
	} else {
		mx_cpu_delay_type = MX_CPU_USE_XCHG;
	}

	if ( (cpu_family == 3) || (cpu_family == 4) ) {
		mx_delay_instructions_per_microsecond = cpu_mhz / 3.0;
	} else {
		mx_delay_instructions_per_microsecond = cpu_mhz;
	}

	return;

}

#elif defined(OS_WIN32)

/******* Microsoft Win32 *******/

#include <windows.h>

static int mx_udelay_init_invoked = FALSE;

static int mx_performance_counters_are_available = FALSE;

static double mx_performance_counter_ticks_per_microsecond = 0.0;

static void
mx_udelay_error_message( const char *function_name )
{
	static const char fname[] = "mx_udelay_error_message()";

	DWORD last_error_code;
	TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

	last_error_code = GetLastError();

	mx_win32_error_message( last_error_code,
		message_buffer, sizeof(message_buffer) );

	(void) mx_error( MXE_FUNCTION_FAILED, fname,
	"Error invoking %s.  Win32 error code = %ld, error message = '%s'",
		function_name, last_error_code, message_buffer );

	return;
}

MX_EXPORT void
mx_udelay( unsigned long microseconds )
{
	if ( mx_udelay_init_invoked == FALSE )
		mx_udelay_init();

	if ( mx_performance_counters_are_available == FALSE ) {
		mx_usleep( microseconds );
	} else {
		LARGE_INTEGER counter_value;
		LONGLONG counter_start_value, counter_end_value;
		LONGLONG counter_current_value, counter_delay_ticks;
		BOOL status;

		counter_delay_ticks = (LONGLONG)
	( mx_performance_counter_ticks_per_microsecond * (double)microseconds );

		status = QueryPerformanceCounter( &counter_value );

		if ( status == 0 ) {
			mx_udelay_error_message( "QueryPerformanceCounter()" );
			return;
		}

		counter_start_value = counter_value.QuadPart;

		counter_end_value = counter_start_value + counter_delay_ticks;

		/* Sit in a loop until we reach the end of the delay time. */

		counter_current_value = counter_start_value;

		while ( counter_current_value < counter_end_value ) {

			status = QueryPerformanceCounter( &counter_value );

			if ( status == 0 ) {
				mx_udelay_error_message(
					"QueryPerformanceCounter()" );
				return;
			}

			counter_current_value = counter_value.QuadPart;
		}
	}
	return;
}

MX_EXPORT void
mx_udelay_init( void )
{
	LARGE_INTEGER performance_frequency;
	BOOL status;

	mx_udelay_init_invoked = TRUE;

	status = QueryPerformanceFrequency( &performance_frequency );

	if ( status == 0 ) {
		mx_udelay_error_message( "QueryPerformanceFrequency()" );

		mx_performance_counters_are_available = FALSE;

		mx_warning(
			"Win32 performance counters are not available, "
			"so we will revert to using mx_usleep() instead." );
	} else {
		mx_performance_counters_are_available = TRUE;

		mx_performance_counter_ticks_per_microsecond
			= 1.0e-6 * (double) performance_frequency.QuadPart;
	}
	return;
}

#else

/******* None of the above ********/

static int mx_udelay_init_invoked = FALSE;

MX_EXPORT void
mx_udelay( unsigned long microseconds )
{
	if ( mx_udelay_init_invoked == FALSE )
		mx_udelay_init();

	mx_usleep( microseconds );
}

MX_EXPORT void
mx_udelay_init( void )
{
	mx_udelay_init_invoked = TRUE;

	mx_warning(
		"mx_udelay() is not yet implemented on this operating system.  "
		"We will revert to using mx_usleep() instead." );
	return;
}

#endif
